;;;
;;; charset.scm
;;; scsh-like character sets
;;;
;;; Copyright (C) 1998-1999, Ivan Demakov.
;;;
;;; Permission to use, copy, modify, and distribute this software and its
;;; documentation for any purpose, without fee, and without a written
;;; agreement is hereby granted, provided that the above copyright notice
;;; and this paragraph and the following two paragraphs appear in all copies.
;;; Modifications to this software may be copyrighted by their authors
;;; and need not follow the licensing terms described here, provided that
;;; the new terms are clearly indicated on the first page of each file where
;;; they apply.
;;;
;;; IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
;;; FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
;;; INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
;;; DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
;;; POSSIBILITY OF SUCH DAMAGE.
;;;
;;; THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
;;; INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
;;; AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
;;; ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
;;; TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
;;;
;;;
;;; Author:        Ivan Demakov <demakov@users.sourceforge.net>
;;; Creation date: Thu Jun 25 06:52:52 1998
;;; Last Update:   $Author: demakov $ $Date: 2000/07/25 19:07:15 $
;;;
;;;
;;; $Id: charset.scm,v 1.1.1.1 2000/07/25 19:07:15 demakov Exp $
;;;
;;;
;;; Comments:
;;;
;;;

(define (char->ascii x)
  (char->integer x))

(define (ascii->char x)
  (integer->char x))


(define (make-char-set)
  (vector '<char-set> 0))

(define (char-set-add x i)
  (vector-set! x 1 (logior (vector-ref x 1) (logash 1 i))))

(define (char-set-get x i)
  (logbit? (vector-ref x 1) i))

(define (char-set? x)
  (and (vector? x)
       (= (vector-length x) 2)
       (eq? (vector-ref x 0) '<char-set>)))


(define (char-set . chars)
  (chars->char-set chars))

(define (chars->char-set chars)
  (let ((s (make-char-set)))
    (for-each (lambda (char)
		(char-set-add s (char->ascii char)))
	      chars)
    s))

(define (string->char-set str)
  (let ((s (make-char-set)))
    (do ((i (- (string-length str) 1) (- i 1)))
	((< i 0) s)
      (char-set-add s (char->ascii (string-ref str i))))))

(define (ascii-range->char-set lower upper)
  (let ((s (make-char-set)))
    (do ((code lower (+ code 1)))
	((>= code upper) s)
      (char-set-add s code))))

(define (predicate->char-set predicate)
  (let ((s (make-char-set)))
    (do ((code 0 (+ code 1)))
	((>= 256 upper) s)
      (if (predicate (ascii->char code))
	  (char-set-add s code)))))


;;; {string, char, char-set, char predicate} -> char-set

(define (->char-set x)
  (cond ((char-set? x) x)
	((string? x) (string->char-set x))
	((char? x) (char-set x))
	((procedure? x) (predicate->char-set x))
	(else
	 (error "->char-set: Not a charset, string, char, or predicate: ~s"
		x))))


(define (char-set-members char-set)
  (let loop ((code 0) (res '()))
    (cond ((>= code 256) res)
	  ((char-set-get char-set code)
	   (loop (+ code 1) (cons (ascii->char code) res)))
	  (else (loop (+ code 1) res)))))


(define (char-set-contains? char-set char)
  (char-set-get char-set (char-ascii char)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (char-set-invert char-set)
  (predicate->char-set
   (lambda (char) (not (char-set-contains? char-set char)))))

(define (char-set-union char-set-1 char-set-2)
  (predicate->char-set
   (lambda (char)
     (or (char-set-contains? char-set-1 char)
	 (char-set-contains? char-set-2 char)))))

(define (char-set-intersection char-set-1 char-set-2)
  (predicate->char-set
   (lambda (char)
     (and (char-set-contains? char-set-1 char)
	  (char-set-contains? char-set-2 char)))))

(define (char-set-difference char-set-1 char-set-2)
  (predicate->char-set
   (lambda (char)
     (and (char-set-contains? char-set-1 char)
	  (not (char-set-contains? char-set-2 char))))))


;;;; System Character Sets

(define char-set:upper-case (ascii-range->char-set #x41 #x5B))
(define char-set:lower-case (ascii-range->char-set #x61 #x7B))
(define char-set:numeric (ascii-range->char-set #x30 #x3A))
(define char-set:graphic (ascii-range->char-set #x20 #x7F))
(define char-set:not-graphic (char-set-invert char-set:graphic))
(define char-set:whitespace
  (char-set char:newline char:tab char:linefeed 
	    char:page char:return char:space))
(define char-set:not-whitespace (char-set-invert char-set:whitespace))
(define char-set:alphabetic
  (char-set-union char-set:upper-case char-set:lower-case))
(define char-set:alphanumeric
  (char-set-union char-set:alphabetic char-set:numeric))


;;; End of code
